// Ex0703: フォトリフレクタの自動校正
#include <FastLED.h>

// LED 16個
const int DATA_PIN = 19; // LEDの接続ピン
const int NUM_LEDS = 16; // LEDの個数
CRGB leds[NUM_LEDS]; // LEDの配列

// フォトリフレクタ 5個
#define NUM_PR  5 // センサの数
const int pr_pins[NUM_PR] = {A3, A6, A7, A4, A5}; // ピン定義
int pr[NUM_PR] = {0};
int pr_bin[NUM_PR] = {0}; // ラインの有無
int pr_min[NUM_PR] = {4095, 4095, 4095, 4095, 4095}; // 最小値
int pr_max[NUM_PR] = {0}; // 最大値
float pr_nrm[NUM_PR] = {0}; // 正規化0.0-1.0
unsigned long pr_time = 0; // フォトインタラプタ用タイマー計測値

// モーターの変数
const int MOTOR_L_CWCCW = 16; // APH direction
const int MOTOR_R_CWCCW = 17; // BPH direction
const int MOTOR_L_PWM = 26; // AEN pwm
const int MOTOR_R_PWM = 4; // BEN pwm

// PWMの設定
const int CH_0 = 0;
const int CH_1 = 1;
const int LEDC_TIMER_BIT = 10;   // PWMの範囲(10bit: 0-1023)
const int LEDC_BASE_FREQ = 10000; // 周波数(Hz) 100khz
const int VALUE_MAX = 1023;      // PWMの最大値
const int sp  = 300; // モーターの基本スピード: 0 - 1023

// スイッチの変数
const int SW_PIN = 25;

void setup() {
  Serial.begin(115200);
  delay(100);

  // モータのポート設定
  pinMode(MOTOR_L_CWCCW, OUTPUT);
  pinMode(MOTOR_R_CWCCW, OUTPUT);

  // PWM設定
  ledcSetup(CH_0, LEDC_BASE_FREQ, LEDC_TIMER_BIT);
  ledcSetup(CH_1, LEDC_BASE_FREQ, LEDC_TIMER_BIT);
  ledcAttachPin(MOTOR_L_PWM, CH_0);
  ledcAttachPin(MOTOR_R_PWM, CH_1);
  ledcWrite(CH_0, 0);
  ledcWrite(CH_1, 0);

  // LED設定
  FastLED.addLeds<WS2812B, DATA_PIN, GRB>(leds, NUM_LEDS);
  FastLED.setBrightness(32); // 最大の明るさ 0-255

  // スイッチのポート設定
  pinMode(SW_PIN, INPUT_PULLUP);

  while (digitalRead(SW_PIN) == HIGH); // スイッチが押されるまで待機

  // 自動校正開始
  pr_calib(); // フォトリフレクタの自動校正

  while (digitalRead(SW_PIN) == HIGH); // スイッチが押されるまで待機
}

void loop() {
  if (millis() - pr_time > 10) { // 10ms毎に実行
    pr_time = millis();
    get_pr(); // フォトリフレクタのセンサ値更新
    led_pos(); // センサの反応位置を点灯
    int pos = pr_pos(); // フォトリフレクタの位置(100-500)
//    Serial.println(pos);
    //pr_print_bin();
  }
}

//===========================================
// フォトリフレクタの関数
//===========================================
// フォトリフレクタの位置を100～500で返す
// ライン外:0
// センター：300
// LEFT-----------CENTER-----------RIGHT
// 100,150,200,250,<300>,350,400,450,500
// OUT of LINE: 0
int pr_pos() {
  int pos; // センサの反応位置
  static int last_pos; // 過去の反応位置

  // フォトリフレクタ毎の重み付け値の計算
  int sum = 0, n = 0;
  for (int i = 0; i < NUM_PR; i++) {
    sum += pr_bin[i] * ((i + 1) * 100);
    if (pr_bin[i] == 1) n++;
  }

  if (n > 0) { // 1つ以上反応していたら
    pos = sum / n;
    last_pos = pos; // 1つ過去の値を記録
  } else { // 反応なし
    pos = last_pos; // 過去の値を使用
  }

  return pos;
}

void pr_minmax() {
  for (int i = 0; i < NUM_PR; i++) {
    pr[i] = analogRead(pr_pins[i]);
    if (pr[i] < pr_min[i]) pr_min[i] = pr[i];
    if (pr[i] > pr_max[i]) pr_max[i] = pr[i];
  }
}

// フォトリフレクタの反応(0,1)にあわせてLEDを点灯
void led_pos() {
  for (int i = 0; i < NUM_PR; i++) {
    if (pr_bin[i] == 1) { // フォトリフレクタが反応していたら
      leds[i + 10] = CRGB(255, 255, 0); // 黄色
    } else {
      leds[i + 10] = CRGB(0, 0, 0); // 消灯
    }
  }
  FastLED.show(); // 表示を更新
}

// フォトリフレクタの値取得，最小・最大を調べ，閾値処理
void get_pr() {
  // センサの値を取得
  for (int i = 0; i < NUM_PR; i++) {
    pr[i] = analogRead(pr_pins[i]);
  }
  // 最小，最大
  for (int i = 0; i < NUM_PR; i++) {
    if (pr[i] < pr_min[i]) pr_min[i] = pr[i];
    if (pr[i] > pr_max[i]) pr_max[i] = pr[i];
  }
  // 閾値処理
  for (int i = 0; i < NUM_PR; i++) {
    pr_nrm[i] = (pr[i] - pr_min[i]) / (float)(pr_max[i] - pr_min[i]);
    if (pr_nrm[i] > 0.6) pr_bin[i] = 1;
    else pr_bin[i] = 0;
  }
}

// フォトリフレクタの反応(0,1)を表示
void pr_print_bin() {
  for (int i = 0; i < NUM_PR; i++) {
    Serial.print(pr_bin[i]);
    if (i < NUM_PR - 1) Serial.print(",");
  }
  Serial.println();
}

// フォトリフレクタの自動校正
// ラインをトレースして，最大値，最小値を確認
void pr_calib() {
  for (int i = 0; i < 80; i++) {
    if ( (i > 10 && i <= 30) || (i > 50 && i <= 70) ) {
      motor(sp, sp, HIGH, LOW); // 右回転
    } else {
      motor(sp, sp, LOW, HIGH); // 左回転
    }
    pr_minmax();
    delay(40);
  }
  motor(0, 0, LOW, LOW); // 停止
}

//===========================================
// モーターの関数
//===========================================
void motor(int left, int right, int left_c, int right_c) {
  digitalWrite(MOTOR_L_CWCCW, left_c);
  digitalWrite(MOTOR_R_CWCCW, right_c);
  ledcWrite(CH_0, left);
  ledcWrite(CH_1, right);
}
