// Ex0704: ロボットが障害物との距離を一定に保つ(P制御)
// 実行時間, sp, Kp を変えてみる
#include <Wire.h>
#include <VL53L0X.h>
#include <FastLED.h>

// 距離センサ
VL53L0X sensor;
unsigned long dist_t = 0; // タイマー

// LED 16個
const int DATA_PIN = 19; // LEDの接続ピン
const int NUM_LEDS = 16; // LEDの個数
CRGB leds[NUM_LEDS]; // LED用の配列

// モーターの変数
const int MOTOR_L_CWCCW = 16; // APH direction
const int MOTOR_R_CWCCW = 17; // BPH direction
const int MOTOR_L_PWM = 26; // AEN pwm
const int MOTOR_R_PWM = 4; // BEN pwm

// PWM
const int CH_0 = 0;
const int CH_1 = 1;
const int LEDC_TIMER_BIT = 10;   // PWMの範囲(8bit:0-255、10bit:0-1023)
const int LEDC_BASE_FREQ = 100000; // 周波数(Hz) 100khz
const int VALUE_MAX = 1023;      // PWMの最大値
int sp = 300; // モータースピード

// 制御関係
const float KP = 10.0; // 比例定数
float dt = 100; // 目標距離 100[mm]
int r = 0; // 制御量

// スイッチの変数
const int SW_PIN = 25;

void setup() {
  // シリアルポートの設定
  Serial.begin(115200); // bps for bluetooth communication
  delay(100);

  // モータのポート設定
  pinMode(MOTOR_L_CWCCW, OUTPUT);
  pinMode(MOTOR_R_CWCCW, OUTPUT);

  // PWM設定
  ledcSetup(CH_0, LEDC_BASE_FREQ, LEDC_TIMER_BIT);
  ledcSetup(CH_1, LEDC_BASE_FREQ, LEDC_TIMER_BIT);
  ledcAttachPin(MOTOR_L_PWM, CH_0);
  ledcAttachPin(MOTOR_R_PWM, CH_1);
  ledcWrite(CH_0, 0);
  ledcWrite(CH_1, 0);

  Wire.begin();
  sensor.init();
  sensor.setTimeout(500);
  sensor.startContinuous();

  // スイッチピン設定
  pinMode(SW_PIN, INPUT_PULLUP);

  // スイッチが押されるまで待機
  while (digitalRead(SW_PIN) == HIGH);
}

void loop() {
  if (millis() - dist_t > 10) { // 10ms毎に更新
    dist_t = millis();
    int d = dist_sensor(); // 距離計測 1ms
    int s = robot_state(d); // ロボットの状態 <1ms
    robot_act(s); // ロボットの動作
  }
}

int robot_state(int d) {
  int s;

  r = KP * (dt - (float)d); // 比例制御
  if (r > sp) r = sp; // 設定速度を超過
  if (r < -sp) r = -sp;  // 設定速度を超過
  //  Serial.println(r);

  if (r == 0 || d == -1) {
    s = 0; // 停止
  }
  else if (r > 0) {
    s = 1; // 後退
  }
  else if (r < 0) {
    s = 2; // 前進
  }

  return s;
}

void robot_act(int s) {
  switch (s) {
    case 0: // 停止
      motor(0, 0, HIGH, HIGH);
      break;
    case 1: // 後退
      motor((int)r, (int)r, LOW, LOW);
      break;
    case 2: // 前進
      motor((int) - r, (int) - r, HIGH, HIGH);
      break;
    default: // 停止
      motor(0, 0, HIGH, HIGH);
      break;
  }
}

//===========================================
// 距離の計測
// 関数: getDistance
// 戻り値：距離[mm]
//===========================================
int dist_sensor() {
  unsigned int d = sensor.readRangeContinuousMillimeters();
  //  Serial.println(d);
  if (d >= 0 && d < 500) {
    return (int)d;
  } else {
    return -1;
  }
}

//===========================================
// モーターの関数
//===========================================
void motor(int left, int right, int left_c, int right_c) {
  digitalWrite(MOTOR_L_CWCCW, left_c);
  digitalWrite(MOTOR_R_CWCCW, right_c);
  ledcWrite(CH_0, left);
  ledcWrite(CH_1, right);
}
