// Ex0706: ライントレース + 自動校正(EEPROM)
#include <FastLED.h>
#include <EEPROM.h> // 校正値(最小，最大)をEEPROMに保存

// LED 16bit
const int DATA_PIN = 19;
const int NUM_LEDS = 16;
CRGB leds[NUM_LEDS];

// P制御用
float KP = 4.0; // Kp * +-100
#define LIMIT 600

// モーターの変数
const int MOTOR_L_CWCCW = 16; // APH direction
const int MOTOR_R_CWCCW = 17; // BPH direction
const int MOTOR_L_PWM = 26; // AEN pwm
const int MOTOR_R_PWM = 4; // BEN pwm

// PWM
const int CH_0 = 0;
const int CH_1 = 1;
const int LEDC_TIMER_BIT = 10;   // PWMの範囲(10bit: 0-1023)
const int LEDC_BASE_FREQ = 10000; // 周波数(Hz) 100khz
const int VALUE_MAX = 1023;      // PWMの最大値
const int sp  = 300; // モーターの基本スピード: 0 - 1023

#define NUM_PR  5 // センサの数
const int pr_pins[NUM_PR] = {A3, A6, A7, A4, A5}; // ピン定義
int pr[NUM_PR] = {0};
int pr_bin[NUM_PR] = {0}; // ラインの有無
int pr_min[NUM_PR] = {4095, 4095, 4095, 4095, 4095}; // 最小値
int pr_max[NUM_PR] = {0}; // 最大値
float pr_nrm[NUM_PR] = {0}; // 正規化0.0-1.0
unsigned long pr_time = 0; // フォトインタラプタ用タイマー計測値
const int pos_t = 300; // 目標値

// スイッチの変数
const int SW_PIN = 25;
int sw_calib = 0;

void setup() {
  // シリアルポートの設定
  Serial.begin(115200); // bps for bluetooth communication
  delay(100);

  // EEPROM
  EEPROM.begin(100); // EEPROMに100byteの変数用の領域確保

  // モータのポート設定
  pinMode(MOTOR_L_CWCCW, OUTPUT);
  pinMode(MOTOR_R_CWCCW, OUTPUT);

  // PWM
  ledcSetup(CH_0, LEDC_BASE_FREQ, LEDC_TIMER_BIT);
  ledcSetup(CH_1, LEDC_BASE_FREQ, LEDC_TIMER_BIT);
  ledcAttachPin(MOTOR_L_PWM, CH_0);
  ledcAttachPin(MOTOR_R_PWM, CH_1);
  ledcWrite(CH_0, 0);
  ledcWrite(CH_1, 0);

  // led
  FastLED.addLeds<WS2812B, DATA_PIN, GRB>(leds, NUM_LEDS);
  FastLED.setBrightness(32); // 最大の明るさ 0-255

  // スイッチ
  pinMode(SW_PIN, INPUT_PULLUP);

  // 自動校正の有無の確認
  auto_calib();

  // スイッチを押すまで待機
  while (digitalRead(SW_PIN) == HIGH);
  pr_print_minmax();
}

void loop() {
  if (millis() - pr_time > 10) { // 10ms毎に更新
    pr_time = millis();
    get_pr(); // フォトリフレクタのセンサ値更新
    led_pos();
    int pos = pr_pos(); // フォトリフレクタの位置(100-500)
    pd(pos); // モーター制御
  }
}


//===========================================
// 自動校正の関数
//===========================================
void auto_calib() {
  int sw_calib = 0; // 自動校正の有無

  // 起動時に校正の有無を確認
  if (digitalRead(SW_PIN) == LOW) { // スイッチが押されていたら
    sw_calib = 1; // 校正をON
    for (int i = 0; i < 3; i++) { // 3回白色点滅
      leds[0] = CRGB(255, 255, 255);
      FastLED.show();
      delay(200);
      leds[0] = CRGB(0, 0, 0);
      FastLED.show();
      delay(200);
    }
  }

  if (sw_calib == 1) {  // 自動校正開始
    sw_calib = 0;
    pr_calib(); // フォトリフレクタの自動校正
    eeprom_write(); // EEPROMに校正値を書き込み
  } else {
    eeprom_read(); // EEPROMから校正値を読み込み
  }
}

// EEPROMに最小・最大値の書き込み
void eeprom_write() {
  int n = 0;
  for (int i = 0; i < NUM_PR; i++) {
    EEPROM.put(n, pr_min[i]);
    EEPROM.put(n + 4 * NUM_PR, pr_max[i]);
    n += 4;
  }
  EEPROM.commit(); // EEPROMに書き込み
}

// EEPROMより最小・最大値の読み込み
void eeprom_read() {
  int n = 0;
  for (int i = 0; i < NUM_PR; i++) {
    EEPROM.get(n, pr_min[i]);
    EEPROM.get(n + 4 * NUM_PR, pr_max[i]);
    n += 4;
  }
  EEPROM.end();
}

//===========================================
// モーター制御の関数
//===========================================
void pd(int pos) {
  int error = pos_t - pos; // 目標値と計測位置の差
  int sp_diff = KP * error; // 比例制御：KP*error
  int sp_l = sp - sp_diff; // 左モーターの速度
  int sp_r = sp + sp_diff; // 右モーターの速度

  // 左，右モーターの最小値，最大値を超えないか確認
  if (sp_l < 0)   sp_l = 0;
  if (sp_r < 0)   sp_r = 0;
  if (sp_l > LIMIT) sp_l = LIMIT;
  if (sp_r > LIMIT) sp_r = LIMIT;

  motor(sp_l, sp_r, HIGH, HIGH);
}

//===========================================
// フォトリフレクタの関数
//===========================================
// フォトリフレクタの位置を100～500で返す
// ライン外:0
// センター：300
// LEFT-----------CENTER-----------RIGHT
// 100,150,200,250,<300>,350,400,450,500
// OUT of LINE: 0
int pr_pos() {
  int pos;
  static int last_pos;

  int sum = 0, n = 0;
  for (int i = 0; i < NUM_PR; i++) {
    sum += pr_bin[i] * ((i + 1) * 100);
    if (pr_bin[i] == 1) n++;
  }

  if (n > 0) { // ラインの上
    pos = sum / n;
    last_pos = pos;
  } else { // n=0: ラインを見失った
    pos = last_pos;
  }

  return pos;
}

void led_pos() {
  for (int i = 0; i < NUM_PR; i++) {
    if (pr_bin[i] == 1) {
      leds[i + 10] = CRGB(255, 255, 0); // 黄色
    } else {
      leds[i + 10] = CRGB(0, 0, 0); // 消灯
    }
  }
  FastLED.show(); // 表示を更新
}

// フォトリフレクタのセンサ値 pr[]: 0-4095
// フォトリフレクタの正規化値 pr_nrm[]: 0.0-1.0
// フォトリフレクタの2値化 pr_bin[]: 0, 1
void get_pr() {
  for (int i = 0; i < NUM_PR; i++) {
    pr[i] = analogRead(pr_pins[i]);
  }
  for (int i = 0; i < NUM_PR; i++) {
    if (pr[i] < pr_min[i]) pr_min[i] = pr[i];
    if (pr[i] > pr_max[i]) pr_max[i] = pr[i];
    pr_nrm[i] = (pr[i] - pr_min[i]) / (float)(pr_max[i] - pr_min[i]);
    if (pr_nrm[i] > 0.6) pr_bin[i] = 1;
    else pr_bin[i] = 0;
  }
}

void pr_minmax() {
  for (int i = 0; i < NUM_PR; i++) {
    pr[i] = analogRead(pr_pins[i]);
  }
  for (int i = 0; i < NUM_PR; i++) {
    if (pr[i] < pr_min[i]) pr_min[i] = pr[i];
    if (pr[i] > pr_max[i]) pr_max[i] = pr[i];
  }
}

void pr_print_minmax() {
  Serial.print("int pr_min[PR_NUM] = {");
  for (int i = 0; i < NUM_PR; i++) {
    Serial.print(pr_min[i]);
    if (i < NUM_PR - 1) Serial.print(",");
  }
  Serial.println("};");

  Serial.print("int pr_max[PR_NUM] = {");
  for (int i = 0; i < NUM_PR; i++) {
    Serial.print(pr_max[i]);
    if (i < NUM_PR - 1) Serial.print(",");
  }
  Serial.println("};");
}

void pr_calib() {
  for (int i = 0; i < 80; i++) {
    if ( (i > 10 && i <= 30) || (i > 50 && i <= 70) ) {
      motor(sp, sp, HIGH, LOW);
    } else {
      motor(sp, sp, LOW, HIGH);
    }
    pr_minmax();
    delay(40);
  }
  motor(0, 0, LOW, LOW);
}

//===========================================
// モーターの関数
//===========================================
void motor(int left, int right, int left_c, int right_c) {
  digitalWrite(MOTOR_L_CWCCW, left_c);
  digitalWrite(MOTOR_R_CWCCW, right_c);
  ledcWrite(CH_0, left);
  ledcWrite(CH_1, right);
}
